from __future__ import division

'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "lib"
directory.
'''
import re
import random

import pygame
from pygame.locals import *
from pygame.sprite import RenderClear

import data
from rules import check_rule, apply_rule
from world import create_world
from helper import OrderedRenderUpdates
from gui import prompt, message


ARROWS = {pygame.K_LEFT: False, pygame.K_RIGHT: False, pygame.K_UP: False, pygame.K_DOWN: False}


def main():
    pygame.init()
    screen = pygame.display.set_mode((640, 480))
    pygame.display.set_caption('Galvinball!')

    # Load background.
    background = pygame.image.load(data.load("background.png")).convert()
    screen.blit(background, (0, 0))
    pygame.display.flip()

    # Create world.
    level = 1
    player, objects, groups, allsprites = create_world(level)
    winners = []
    rules = ['WHOEVER IS OVER BALL WINS']

    # Load music.
    pygame.mixer.music.load('data/spotty.ogg')
    pygame.mixer.music.play(-1)

    clock = pygame.time.Clock()
    while 1:
        clock.tick(20)

        # Init all object actions.
        for object in objects:
            object._init()
            
        # Winner?
        for object in objects:
            if getattr(object, 'winner', False):
                winners.append(object)
            # Announce winner!
            if winners:
                # show message.
                winners = [getattr(w, 'name', w.group) for w in winners]
                text = (', '.join(winners[:-1]) + ' and ' + winners[-1] + ' won!').lstrip(' and ').upper()
                message(text, screen)
                if getattr(player, 'winner', False):
                    if level == 10:
                        message("YOU'RE THE KING OF GALVINBALL! KTHXBY!", screen, y=240)
                        return
                    level += 1
                player, objects, groups, allsprites = create_world(level)
                winners = []
                rules = ['WHOEVER IS OVER BALL WINS']
                screen.blit(background, (0, 0))
                pygame.display.flip()

        for event in pygame.event.get():
            if event.type == QUIT:
                return
            elif event.type == pygame.KEYUP:
                ARROWS[event.key] = False

            elif event.type == pygame.KEYDOWN:
                ARROWS[event.key] = True

                if event.key == pygame.K_RETURN:
                    new_rule = prompt(screen, y=480-32)
                    if new_rule.strip():
                        # Check that rule is in uppercase...
                        if new_rule.upper() != new_rule or player.mute:
                            if not computer.mute:
                                message("SAY IT LOUDER, I CAN'T HEAR YOU!", screen)
                            else:
                                message("(...)", screen)
                        # Add the rule...
                        elif check_rule(new_rule):
                            rules.append(new_rule)
                        # ... or say that we didn't understand it.
                        else:
                            message("WHAT? I COULDN'T UNDERSTAND THAT...", screen)
                    screen.blit(background, (0, 0))
                    pygame.display.flip()

                elif event.key == pygame.K_r:
                    player, objects, groups, allsprites = create_world(level)
                    winners = []
                    rules = ['WHOEVER IS OVER BALL WINS']
                    screen.blit(background, (0, 0))
                    pygame.display.flip()

        # computer generated rules.
        computers = [o for o in objects if getattr(o, 'name', None) == 'Suzanne']
        for computer in computers:
            if not computer.mute and computer.rules and random.random() < 0.001:
                new_rule = computer.rules.pop()
                if check_rule(new_rule):
                    message(new_rule, screen)
                    rules.append(new_rule)
                    screen.blit(background, (0, 0))
                    pygame.display.flip()
                else:
                    print 'Broken rule: %s' % new_rule

        # Apply rules.
        for rule in rules:
            apply_rule(rule, objects, groups)

        # Move player.
        if ARROWS[pygame.K_LEFT]:
            player.direction = (-1, 0)
        elif ARROWS[pygame.K_RIGHT]:
            player.direction = (1, 0)
        elif ARROWS[pygame.K_UP]:
            player.direction = (0, -1)
        elif ARROWS[pygame.K_DOWN]:
            player.direction = (0, 1)

        allsprites.clear(screen, background)
        allsprites.update(groups)
        dirty = allsprites.draw(screen)
        pygame.display.update(dirty)

        # Finalize actions.
        for object in objects:
            object._finish()
